import pytest

from kotsu import error, registration


def fake_entity_factory(param_1, param_2):
    assert param_1 == "PARAM_1"
    assert param_2 == "PARAM_2"


class FakeEntity:
    def __init__(self, param_1, param_2):
        assert param_1 == "PARAM_1"
        assert param_2 == "PARAM_2"


@pytest.mark.parametrize(
    "entry_point",
    [
        fake_entity_factory,
        FakeEntity,
        "tests.test_registration:fake_entity_factory",
        "tests.test_registration:FakeEntity",
    ],
)
def test_registration(entry_point):
    registry = registration._Registry()

    registry.register(
        id="SomeEntity-v0",
        entry_point=entry_point,
        kwargs={"param_1": "PARAM_1", "param_2": "PARAM_2"},
    )

    registry.register(
        id="SomeEntity-v1",
        entry_point=entry_point,
        kwargs={"param_1": "WRONG_PARAM_1", "param_2": "PARAM_2"},
    )

    # We can't use mocks, since we want to test importing via strings
    # So we run the asserts in the function
    # But we protect against the function never being run (and the test silently passing)
    # by also asserting that there _is_ a failure if we run with the wrong parameters
    registry.make(id="SomeEntity-v0")
    with pytest.raises(AssertionError):
        registry.make(id="SomeEntity-v1")

    all_specs = list(registry.all())
    assert len(all_specs) == 2
    assert all_specs[0].id == "SomeEntity-v0"
    assert all_specs[1].id == "SomeEntity-v1"


def test_register_duplicate_id():
    registry = registration._Registry()

    registry.register("Entity-v0", "fake_entry_point")
    with pytest.warns(UserWarning, match=r"Entity with ID \[id=Entity-v0] already registered.*"):
        registry.register("Entity-v0", "fake_entry_point")


def test_make_missing_entity():
    registry = registration._Registry()

    with pytest.raises(KeyError, match=r"No registered entity with ID"):
        registry.make("Entity-v0")


def test_make_deprecated_entity():
    registry = registration._Registry()

    registry.register("Entity-v0", "fake_entry_point", deprecated=True)
    with pytest.raises(error.EntityIsDeprecated, match=r"Attempting to make deprecated entity"):
        registry.make("Entity-v0")


@pytest.mark.parametrize(
    "good_id",
    [
        "user1/entity_name_1-v1",
        "user1/entity_name_1_{param_1=10}_{param_2=relu}-v1",
        "user1/entity_name_1-v12",
        "user1/entity_name_1-v1.2.0",
        "user1/entity:name.1_[param_1=10]_{param_2=relu}-v1",
    ],
)
def test_well_formed_entity_id(good_id):
    registry = registration._Registry()
    registry.register(good_id, "fake_entry_point")


@pytest.mark.parametrize(
    "bad_id",
    [
        "EntityMissingVersionName",
        "EntityNoDashBetweenVersionv0",
        "user1/user2/EntityTwoUsers-v0",
        "EntityNaughty$ymbols-v0",
    ],
)
def test_malformed_entity_id(bad_id):
    registry = registration._Registry()

    with pytest.raises(ValueError, match=r"Attempted to register malformed entity ID"):
        registry.register(bad_id, "fake_entry_point")



"""Interface for running a registry of models on a registry of validations."""
from typing import List, Optional, Tuple, Union
from typing_extensions import Literal
from kotsu.typing import Model, Results, Validation

import functools
import logging
import os
import time

import pandas as pd

from kotsu import store
from kotsu.registration import ModelRegistry, ModelSpec, ValidationRegistry, ValidationSpec


logger = logging.getLogger(__name__)


def run(
    model_registry: ModelRegistry,
    validation_registry: ValidationRegistry,
    results_path: str = "./validation_results.csv",
    force_rerun: Optional[Union[Literal["all"], List[str]]] = None,
    artefacts_store_dir: Optional[str] = None,
    run_params: Optional[dict] = None,
):
    """Run a registry of models through a registry of validations.

    Args:
        model_registry: A ModelRegistry containing the registry of models to be run through
            validations.
        validation_registry: A ValidationRegistry containing the registry of validations to run
            each model through.
        results_path: The file path to which the results will be written to, and results from prior
            runs will be read from.
        force_rerun: Argument to force models to rerun on validations. Model-validation
            combinations without results will always be ran, as well as models that are forced via
            this argument, which will overwrite previous results.
            - if `force_rerun` = None, don't force rerun any models (default)
            - if `force_rerun` = "all", force rerun all models
            - if `force_rerun` = list of string ids, force rerun only these specified models
        artefacts_store_dir: A directory path or URI location to store extra output artefacts
            of the validations and models.
            If not None, then validations will be passed two kwargs; `validation_artefacts_dir` and
            `model_artefacts_dir`.
        run_params: A dictionary of optional run parameters.
    """
    if run_params is None:
        run_params = {}

    try:
        results_df = pd.read_csv(results_path)
    except FileNotFoundError:
        results_df = pd.DataFrame(columns=["validation_id", "model_id", "runtime_secs"])
        results_df["runtime_secs"] = results_df["runtime_secs"].astype(int)

    results_df = results_df.set_index(["validation_id", "model_id"], drop=False)
    results_list = []

    for validation_spec in validation_registry.all():
        if validation_spec.deprecated:
            logger.info(f"Skipping validation: {validation_spec.id} - as is deprecated.")
            continue
        for model_spec in model_registry.all():
            if model_spec.deprecated:
                logger.info(f"Skipping model: {model_spec.id} - as is deprecated.")
                continue

            if (
                not force_rerun == "all"
                and not (isinstance(force_rerun, list) and model_spec.id in force_rerun)
                and (validation_spec.id, model_spec.id) in results_df.index
            ):
                logger.info(
                    f"Skipping validation - model: {validation_spec.id} - {model_spec.id}"
                    ", as found prior result in results."
                )
                continue

            logger.info(f"Running validation - model: {validation_spec.id} - {model_spec.id}")

            validation = validation_spec.make()
            validation = _form_validation_partial_with_store_dirs(
                validation,
                artefacts_store_dir,
                validation_spec,
                model_spec,
            )

            model = model_spec.make()
            results, elapsed_secs = _run_validation_model(validation, model, run_params)
            results = _add_meta_data_to_results(results, elapsed_secs, validation_spec, model_spec)
            results_list.append(results)

    additional_results_df = pd.DataFrame.from_records(results_list)
    results_df = results_df.append(additional_results_df, ignore_index=True)
    results_df = results_df.drop_duplicates(subset=["validation_id", "model_id"], keep="last")
    results_df = results_df.sort_values(by=["validation_id", "model_id"]).reset_index(drop=True)
    store.write(
        results_df, results_path, to_front_cols=["validation_id", "model_id", "runtime_secs"]
    )


def _form_validation_partial_with_store_dirs(
    validation: Validation,
    artefacts_store_dir: Union[str, None],
    validation_spec: ValidationSpec,
    model_spec: ModelSpec,
):
    """Form a partial of the validation with formed validation and model artefacts dirs if needed.

    Also makes any needed dirs for the artefacts dirs.
    """
    if artefacts_store_dir is not None:
        validation_artefacts_dir = os.path.join(artefacts_store_dir, f"{validation_spec.id}/")
        model_artefacts_dir = os.path.join(validation_artefacts_dir, f"{model_spec.id}/")
        os.makedirs(model_artefacts_dir, exist_ok=True)
        validation = functools.partial(
            validation,
            validation_artefacts_dir=validation_artefacts_dir,
            model_artefacts_dir=model_artefacts_dir,
        )
    else:
        pass
    return validation


def _add_meta_data_to_results(
    results: Results,
    elapsed_secs: float,
    validation_spec: ValidationSpec,
    model_spec: ModelSpec,
) -> Results:
    """Add meta data to results, raising if keys clash."""
    results_meta_data: Results = {
        "validation_id": validation_spec.id,
        "model_id": model_spec.id,
        "runtime_secs": elapsed_secs,
    }
    if bool(set(results) & set(results_meta_data)):
        raise ValueError(
            (
                f"Validation:{validation_spec.id} on model:{model_spec.id} "
                f"returned results:{results} which contains a privileged key name."
            )
        )
    return {**results, **results_meta_data}


def _run_validation_model(
    validation: Validation, model: Model, run_params: dict
) -> Tuple[Results, float]:
    """Run given validation on given model, and store the results.

    Returns:
        A tuple of (dict of results: Results type, elapsed time in seconds)
    """
    start_time = time.time()
    results = validation(model, **run_params)
    elapsed_secs = time.time() - start_time
    return results, elapsed_secs



"""An end to end example and test."""
from typing import Callable

import pickle

from sklearn import datasets, svm
from sklearn.model_selection import cross_validate

import kotsu


model_registry = kotsu.registration.ModelRegistry()

model_registry.register(
    id="SVC-v1",
    entry_point=svm.SVC,
    kwargs={"kernel": "linear", "C": 1, "random_state": 1},
)

model_registry.register(
    id="SVC-v2",
    entry_point=svm.SVC,
    kwargs={"kernel": "linear", "C": 0.5, "random_state": 1},
)

validation_registry = kotsu.registration.ValidationRegistry()


def factory_iris_cross_val(folds: int) -> Callable:
    """Factory for iris cross validation."""

    def iris_cross_val(model, validation_artefacts_dir=None, model_artefacts_dir=None) -> dict:
        """Iris classification cross validation."""
        X, y = datasets.load_iris(return_X_y=True)
        scores = cross_validate(model, X, y, cv=folds, return_estimator=True)

        if model_artefacts_dir:
            # Save the trained models from each fold
            for fold_idx, model in enumerate(scores["estimator"]):
                with open(model_artefacts_dir + f"model_from_fold_{fold_idx}.pk", "wb") as f:
                    pickle.dump(model, f)

        results = {f"fold_{i}_score": score for i, score in enumerate(scores["test_score"])}
        results["mean_score"] = scores["test_score"].mean()
        results["std_score"] = scores["test_score"].std()
        return results

    return iris_cross_val


validation_registry.register(
    id="iris_cross_val-v1",
    entry_point=factory_iris_cross_val,
    kwargs={"folds": 5},
)

validation_registry.register(
    id="iris_cross_val-v2",
    entry_point=factory_iris_cross_val,
    kwargs={"folds": 10},
)


def test_run(tmpdir):
    kotsu.run.run(
        model_registry, validation_registry, results_path=str(tmpdir) + "/validation_results.csv"
    )


def test_run_save_models(tmpdir):
    kotsu.run.run(
        model_registry,
        validation_registry,
        results_path=str(tmpdir) + "/validation_results.csv",
        artefacts_store_dir=str(tmpdir),
    )



"""Define types and protocols for core interfaces for kotsu.

Mainly for documentation purposes.
"""
from typing import Any, Callable, Dict, Union


# The results from a validation run on a particular model.
# Each Results dict will become one row of a combined results dataframe. Each key within the
# Results dict will become a column of that dataframe, with values for matching keys across Results
# put into the same matching column.
# An example result from a 3-fold cross validation might be:
# `{"average_score": _, "standard_deviation_scores": _, "fold_1_score": _, "fold_2_score"...}`
Results = Dict[str, Union[float, str]]

# Models within kotsu are not constrained in any way. It is up to the user to write Models that
# have an interface compatible with their Validations.
Model = Any

# A Validation is a callable (e.g. function) that takes a Model, and returns a Results.
Validation = Callable[..., Results]

# Validations can also take a Model and two artefacts directory kwargs; `validation_artefacts_dir`
# and`model_artefacts_dir`. Each can be used to write any artefacts/data objects formed by the
# Validation or Model respectively. Passing `model_artefacts_dir` to Models is recommended for them
# to use to store output artefacts (e.g. saving model state, or saving training history) to
# directly.
ValidationWithOutputArtefacts = Callable[[Model, str, str], Results]



"""Functionality for storing validation results."""
from typing import List

import pandas as pd


def write(df: pd.DataFrame, results_path: str, to_front_cols: List[str]):
    """Write the results to the results path."""
    df = df[to_front_cols + [col for col in df.columns if col not in to_front_cols]]
    df.to_csv(results_path, index=False)



import pandas as pd
import pytest

from kotsu import store


@pytest.mark.parametrize(
    "to_front_cols",
    [
        [],
        ["result_b", "result_c"],
    ],
)
def test_write(to_front_cols, tmpdir):
    results = pd.DataFrame(
        [
            {"id": "v1", "result_b": 2, "result_c": 3},
            {"id": "v2", "result_b": 20, "result_c": 30},
        ]
    )
    results_path = str(tmpdir) + "validation_results.csv"
    store.write(results, results_path, to_front_cols)

    df = pd.read_csv(results_path)
    assert len(df) == 2
    assert len(df.columns) == 3
    assert df.loc[0, "id"] == "v1"
    if to_front_cols:
        assert (df.columns[: len(to_front_cols)] == to_front_cols).all()



"""Registration of entities (models, or validations).

Enables registering specifications of entities and parameters combinations, under a unique ID,
which can be passed to kotsu's run interface.

Based on: https://github.com/openai/gym/blob/master/gym/envs/registration.py
"""
from typing import Callable, Generic, Optional, TypeVar, Union
from kotsu.typing import Model, Validation

import importlib
import logging
import re
import warnings

from kotsu import error


logger = logging.getLogger(__name__)


Entity = TypeVar("Entity")

# A unique ID for an entity; a name followed by a version number.
# Entity-name is group 1, version is group 2.
# [username/](entity-name)-v(major).(minor)
# See tests for examples of well formed IDs.
entity_id_re = re.compile(r"^(?:[\w:-]+\/)?([\w:.\-{}=\[\]]+)-v([\d.]+)$")


def _load(name: str):
    """Load a python object from string.

    Args:
        name: of the form `path.to.module:object`
    """
    mod_name, attr_name = name.split(":")
    mod = importlib.import_module(mod_name)
    fn = getattr(mod, attr_name)
    return fn


class _Spec(Generic[Entity]):
    """A specification for a particular instance of an entity.

    Used to register entity and parameters full specification for evaluations.

    Args:
        id: A unique entity ID
            Required format; [username/](entity-name)-v(version)
            [username/] is optional.
        entry_point: The python entrypoint of the entity class. Should be one of:
            - the string path to the python object (e.g.module.name:factory_func, or
              module.name:Class)
            - the python object (class or factory) itself
        deprecated: Flag to denote whether this entity should be skipped in validation runs and
            considered deprecated and replaced by a more recent/better validation/model
        nondeterministic: Whether this entity is non-deterministic even after seeding
        kwargs: The kwargs to pass to the entity entry point when instantiating the entity
    """

    def __init__(
        self,
        id: str,
        entry_point: Union[Callable, str],
        deprecated: bool = False,
        nondeterministic: bool = False,
        kwargs: Optional[dict] = None,
    ):
        self.id = id
        self.entry_point = entry_point
        self.deprecated = deprecated
        self.nondeterministic = nondeterministic
        self._kwargs = {} if kwargs is None else kwargs

        match = entity_id_re.search(id)
        if not match:
            raise ValueError(
                f"Attempted to register malformed entity ID: [id={id}]. "
                f"(Currently all IDs must be of the form {entity_id_re.pattern}.)"
            )

    def make(self, **kwargs) -> Entity:
        """Instantiates an instance of the entity."""
        if self.deprecated:
            raise error.EntityIsDeprecated(
                f"Attempting to make deprecated entity {self.id}. "
                "(HINT: is there a newer registered version of this entity?)"
            )
        _kwargs = self._kwargs.copy()
        _kwargs.update(kwargs)
        if callable(self.entry_point):
            factory = self.entry_point
        else:
            factory = _load(self.entry_point)
        entity = factory(**_kwargs)

        return entity

    def __repr__(self):
        return "Spec({})".format(self.id)


class _Registry(Generic[Entity]):
    """Register an entity by ID.

    IDs should remain stable over time and should be guaranteed to resolve to the same entity
    dynamics (or be desupported).
    """

    def __init__(self):
        self.entity_specs = {}

    def make(self, id: str, **kwargs) -> Entity:
        """Instantiate an instance of an entity of the given ID."""
        logging.info(f"Making new entity: {id} ({kwargs})")
        try:
            return self.entity_specs[id].make(**kwargs)
        except KeyError:
            raise KeyError(f"No registered entity with ID {id}")

    def all(self):
        """Return all the entitys in the registry."""
        return self.entity_specs.values()

    def register(
        self,
        id: str,
        entry_point: Union[Callable, str],
        deprecated: bool = False,
        nondeterministic: bool = False,
        kwargs: Optional[dict] = None,
    ):
        """Register an entity.

        Args:
            id: A unique entity ID
                Required format; [username/](entity-name)-v(version)
                [username/] is optional.
            entry_point: The python entrypoint of the entity class. Should be one of:
                - the string path to the python object (e.g.module.name:factory_func, or
                  module.name:Class)
                - the python object (class or factory) itself
            deprecated: Flag to denote whether this entity should be skipped in validation runs and
                considered deprecated and replaced by a more recent/better validation/model.
            nondeterministic: Whether this entity is non-deterministic even after seeding
            kwargs: The kwargs to pass to the entity entry point when instantiating the entity
        """
        if id in self.entity_specs:
            warnings.warn(
                f"Entity with ID [id={id}] already registered, but the ID is now being used "
                "to register another entity, OVERWRITING previous registered entity.",
                UserWarning,
            )
        self.entity_specs[id] = _Spec(
            id,
            entry_point,
            deprecated=deprecated,
            nondeterministic=nondeterministic,
            kwargs=kwargs,
        )


ModelSpec = _Spec[Model]
ModelRegistry = _Registry[Model]

ValidationSpec = _Spec[Validation]
ValidationRegistry = _Registry[Validation]



import logging
from unittest import mock

import pandas as pd
import pytest

import kotsu


logger = logging.getLogger(__name__)


class FakeRegistry:
    def __init__(self, ids):
        self.ids = ids
        entitys = []
        instances = []
        for id_ in ids:
            entity = mock.Mock()
            entity.id = id_
            entity.deprecated = False
            entitys.append(entity)
            instance = mock.Mock()
            instance.return_value = {}
            entity.make.return_value = instance
            instances.append(instance)
        self.entitys = entitys
        self.instances = instances

    def all(self):
        return self.entitys


def test_form_results(mocker, tmpdir):
    patched_run_validation_model = mocker.patch(
        "kotsu.run._run_validation_model",
        side_effect=[({"test_result": "result"}, 10), ({"test_result": "result_2"}, 20)],
    )
    patched_store_write = mocker.patch("kotsu.store.write")

    models = ["model_1", "model_2"]
    model_registry = FakeRegistry(models)
    validations = ["validation_1"]
    validation_registry = FakeRegistry(validations)

    results_path = str(tmpdir) + "validation_results.csv"
    kotsu.run.run(model_registry, validation_registry, results_path=results_path)

    results_df = pd.DataFrame(
        [
            {
                "validation_id": "validation_1",
                "model_id": "model_1",
                "runtime_secs": 10,
                "test_result": "result",
            },
            {
                "validation_id": "validation_1",
                "model_id": "model_2",
                "runtime_secs": 20,
                "test_result": "result_2",
            },
        ]
    )

    assert patched_run_validation_model.call_count == 2
    pd.testing.assert_frame_equal(
        patched_store_write.call_args[0][0],
        results_df,
    )
    assert patched_store_write.call_args[0][1] == results_path


@pytest.mark.parametrize("artefacts_store_dir", [None, "test_dir/"])
def test_validation_calls(artefacts_store_dir, mocker):
    _ = mocker.patch("kotsu.store.write")

    models = ["model_1", "model_2", "model_3"]
    model_registry = FakeRegistry(models)
    model_registry.entitys[1].deprecated = True
    validations = ["validation_1", "validation_2"]
    validation_registry = FakeRegistry(validations)
    validation_registry.entitys[1].deprecated = True

    kotsu.run.run(
        model_registry,
        validation_registry,
        artefacts_store_dir=artefacts_store_dir,
    )
    if artefacts_store_dir is not None:
        validation_registry.instances[0].assert_has_calls(
            [
                mock.call(
                    model_registry.instances[0],
                    validation_artefacts_dir=f"{artefacts_store_dir}validation_1/",
                    model_artefacts_dir=f"{artefacts_store_dir}validation_1/model_1/",
                ),
                mock.call(
                    model_registry.instances[2],
                    validation_artefacts_dir=f"{artefacts_store_dir}validation_1/",
                    model_artefacts_dir=f"{artefacts_store_dir}validation_1/model_3/",
                ),
            ]
        )
    else:
        validation_registry.instances[0].assert_has_calls(
            [
                mock.call(
                    model_registry.instances[0],
                ),
                mock.call(
                    model_registry.instances[2],
                ),
            ]
        )
    validation_registry.instances[1].assert_not_called()


def test_passing_run_params_to_validation(mocker):
    _ = mocker.patch("kotsu.store.write")

    models = ["model_1"]
    model_registry = FakeRegistry(models)
    validations = ["validation_1"]
    validation_registry = FakeRegistry(validations)

    kotsu.run.run(
        model_registry,
        validation_registry,
        run_params={"test_param": "test_param_value"},
    )
    validation_registry.instances[0].assert_has_calls(
        [mock.call(model_registry.instances[0], test_param="test_param_value")]
    )


@pytest.mark.parametrize("force_rerun", [None, ["model_1"], "all"])
def test_force_rerun(force_rerun, mocker, tmpdir):
    patched_run_validation_model = mocker.patch(
        "kotsu.run._run_validation_model",
        side_effect=[
            ({"test_result": "result_1"}, 10),
            ({"test_result": "result_2"}, 20),
            ({"test_result": "result_3"}, 30),
            ({"test_result": "result_4"}, 40),
        ],
    )

    models = ["model_1", "model_2"]
    model_registry = FakeRegistry(models)
    validations = ["validation_1"]
    validation_registry = FakeRegistry(validations)

    results_path = str(tmpdir) + "validation_results.csv"
    kotsu.run.run(
        model_registry,
        validation_registry,
        results_path=results_path,
        force_rerun=force_rerun,
    )
    out_df = pd.read_csv(results_path)

    results_df = pd.DataFrame(
        [
            {
                "validation_id": "validation_1",
                "model_id": "model_1",
                "runtime_secs": 10,
                "test_result": "result_1",
            },
            {
                "validation_id": "validation_1",
                "model_id": "model_2",
                "runtime_secs": 20,
                "test_result": "result_2",
            },
        ]
    )

    assert patched_run_validation_model.call_count == 2
    pd.testing.assert_frame_equal(out_df, results_df)

    kotsu.run.run(
        model_registry,
        validation_registry,
        results_path=results_path,
        force_rerun=force_rerun,
    )
    out_df = pd.read_csv(results_path)

    if force_rerun is None:
        assert patched_run_validation_model.call_count == 2
        pd.testing.assert_frame_equal(out_df, results_df)

    if isinstance(force_rerun, list) and force_rerun[0] == "model_1":
        results_df = pd.DataFrame(
            [
                {
                    "validation_id": "validation_1",
                    "model_id": "model_1",
                    "runtime_secs": 30,
                    "test_result": "result_3",
                },
                {
                    "validation_id": "validation_1",
                    "model_id": "model_2",
                    "runtime_secs": 20,
                    "test_result": "result_2",
                },
            ]
        )
        assert patched_run_validation_model.call_count == 3
        pd.testing.assert_frame_equal(out_df, results_df)

    if force_rerun == "all":
        results_df = pd.DataFrame(
            [
                {
                    "validation_id": "validation_1",
                    "model_id": "model_1",
                    "runtime_secs": 30,
                    "test_result": "result_3",
                },
                {
                    "validation_id": "validation_1",
                    "model_id": "model_2",
                    "runtime_secs": 40,
                    "test_result": "result_4",
                },
            ]
        )
        assert patched_run_validation_model.call_count == 4
        pd.testing.assert_frame_equal(out_df, results_df)


@pytest.mark.parametrize(
    "result",
    [
        ({"validation_id": "_", "test_result": "result"}, 10),
        ({"model_id": "_", "test_result": "result_2"}, 20),
        ({"runtime_secs": "_", "test_result": "result_3"}, 30),
    ],
)
def test_raise_if_valiation_returns_privilidged_key_name(result, mocker, tmpdir):
    _ = mocker.patch("kotsu.run._run_validation_model", side_effect=[result])
    models = ["model_1"]
    model_registry = FakeRegistry(models)
    validations = ["validation_1"]
    validation_registry = FakeRegistry(validations)

    results_path = str(tmpdir) + "validation_results.csv"

    with pytest.raises(ValueError):
        kotsu.run.run(
            model_registry,
            validation_registry,
            results_path=results_path,
        )



import pathlib

from setuptools import find_packages, setup

import versioneer


REPO_ROOT = pathlib.Path(__file__).parent

with open(REPO_ROOT / "README.md", encoding="utf-8") as f:
    README = f.read()

REQUIREMENTS = ["pandas"]

setup(
    name="kotsu",
    version=versioneer.get_version(),
    description="Lightweight framework for structured and repeatable model validation",
    long_description=README,
    long_description_content_type="text/markdown",
    author="datavaluepeople",
    author_email="opensource@datavaluepeople.com",
    url="https://github.com/datavaluepeople/kotsu",
    license="MIT",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    python_requires=">=3.7",
    cmdclass=versioneer.get_cmdclass(),
)
